@@ -12,10 +12,16 @@ def get_form_spec(form: DjangoForm) -> FormState:
     helper = form.helper
     helper_has_placeholders = hasattr(helper, 'placeholders')
 
+    # we need to process bound fields to act closely to the original Django forms logic
+    # see django.forms.forms.BaseForm.__iter__
+    #     (as an entry point during rendering forms in templates)
+    bound_fields = {f_name: form[f_name] for f_name in form.fields}
+
     fields = {}
     hidden_fields = []
-    for name, field in form.fields.items():
-        widget = field.widget
+    for name, bound_field in bound_fields.items():
+        field = bound_field.field
+        widget = bound_field.field.widget
         widget_type = type(widget)
         field_spec = {
             'name': name,
@@ -46,7 +52,7 @@ def get_form_spec(form: DjangoForm) -> FormState:
 
         if field_spec[error_name]:
             error_value = get_error_message(name, error_name, helper, global_messages)
-            field_spec['errors'][error_name] = error_value
+            field_spec['errors'][error_name] = str(error_value)
         # also see the logic in validators
 
         # attrs
@@ -61,7 +67,7 @@ def get_form_spec(form: DjangoForm) -> FormState:
         # Validators
         field_spec['validators'] = []
         has_custom_validators = hasattr(helper, 'validators')
-        field_name = type(field).__name__
+        field_name = field.dfp_field_name if hasattr(field, 'dfp_field_name') else type(field).__name__
         match field_name:
             case 'CharField':
                 if hasattr(field, 'max_length') and field.max_length:
@@ -80,12 +86,16 @@ def get_form_spec(form: DjangoForm) -> FormState:
                         field_spec['errors'][validator['name']] = str(error_value)
 
         # Soft validators
+        field_spec['soft_validators'] = []
         has_soft_validators = hasattr(helper, 'soft_validators') and name in helper.soft_validators
-        soft_errors = {key: str(value) for key, value in helper.soft_validators[name].items()} \
-                      if has_soft_validators else {}
-        field_spec['soft_validators'] = [key for key, msg in soft_errors.items()] \
-                                        if has_soft_validators else []
-        field_spec['soft_errors'] = soft_errors if has_soft_validators else []
+        if has_soft_validators:
+            for validator in helper.soft_validators[name]:
+                field_spec['soft_validators'].append({
+                    'name': validator['name'],
+                    'value': validator['value'] if 'value' in validator else None,
+                    'inverse': validator['inverse'] if 'inverse' in validator else False,
+                    'message': str(validator['message']),  # a message is required for now
+                })
 
         # attr: type  + specific settings
         widget_name = widget_type.__name__
@@ -93,6 +103,7 @@ def get_form_spec(form: DjangoForm) -> FormState:
             case 'SlugInput':
                 field_spec['type'] = 'slug'
                 field_spec['prefix'] = widget.prefix if hasattr(widget, 'prefix') else ''
+                field_spec['suggestions'] = widget.suggestions
                 default_initial = ''
             case 'TextInput':
                 field_spec['type'] = 'text'
